from statsmodels.nonparametric.kernel_regression import KernelReg
from typing import Dict, Tuple, List
from dovado_rtl.point_evaluation import evaluate, DesignValue, get_metric
from random import randint, shuffle
from dataclasses import dataclass
import dovado_rtl.global_user_selections as gus
import numpy as np


@dataclass
class Example:
    design_point: List[int]
    design_value: DesignValue


examples: List[Example] = []
examples_updated = True


def add_example(example: Example):
    global examples
    global examples_updated
    examples.append(example)
    shuffle(examples)
    examples_updated = True
    print("EXAMPLES: " + str(examples))


def generate_dataset(
    size: int,
    parameters_range: Dict[str, Tuple[int, int]],
    free_parameters: List[str],
):
    for i in range(0, size):
        design_point = []
        for k in free_parameters:
            design_point.append(
                randint(parameters_range[k][0], parameters_range[k][1])
            )
        design_value = evaluate(tuple(design_point))
        add_example(Example(design_point, design_value))


def get_dependent_variable(
    examples: List[Example], metric: Tuple[str, str]
) -> np.ndarray:
    dependent_variable = []
    for example in examples:
        dependent_variable.append(get_metric(example.design_value, metric))
    return np.array(dependent_variable)


def get_independent_variables(
    examples: List[Example], free_parameters: List[str]
) -> np.ndarray:
    independent_variables = []
    for parameter in free_parameters:
        parameter_accumulator = []
        for example in examples:
            parameter_accumulator.append(example.design_point[parameter])
        independent_variables.append(parameter_accumulator)
    return np.array(independent_variables)


estimator = None


def estimate(design_point: List[float], metric: Tuple[str, str]):
    global estimator
    global examples_updated
    if examples_updated:
        independent_variables = get_independent_variables(
            examples, gus.FREE_PARAMETERS
        )
        estimator = KernelReg(
            get_dependent_variable(examples, metric),
            independent_variables,
            "c" * len(independent_variables),
        )
        examples_updated = False
    estimate, _ = estimator.fit(np.array(design_point))
    return estimate[0]



import dovado_rtl.user_input as user_input
from dovado_rtl.config import get_config
import dovado_rtl.vivado_interaction as vivado
import dovado_rtl.frame_handling as frame
from dovado_rtl.genetic_algorithm import optimize
from dovado_rtl.src_parsing import is_to_box, RTL
import dovado_rtl.global_user_selections as gus
import dovado_rtl.point_evaluation as pe
import dovado_rtl.estimation as es
import dovado_rtl.fitness as fit
import yaml
from pathlib import Path


def main():
    # Starting Vivado in the background
    vivado.start()

    SRC_FOLDER = user_input.ask_code_dir()
    TOP_MODULE, TOP_SRC = user_input.ask_top_module(SRC_FOLDER)
    TOP_SUFFIX = Path(SRC_FOLDER + TOP_SRC).suffix
    if TOP_SUFFIX == ".vhd" or TOP_SUFFIX == ".vhdl":
        TOP_LANG = RTL.VHDL
    elif TOP_SUFFIX == ".v":
        TOP_LANG = RTL.VERILOG
    elif TOP_SUFFIX == ".sv":
        TOP_LANG = RTL.SYSTEM_VERILOG
    else:
        raise ValueError(
            "Parsed files must have .v, .vhd, .vhdl or .sv as a suffix not "
            + TOP_SUFFIX
        )
    print("Current Language: " + str(TOP_LANG))
    SYNTHESIS_PART = user_input.ask_part()

    # STOP_STEP can take the following values:
    #  "synthesis", "implementation"
    STOP_STEP = user_input.ask_stop_step()

    TO_BOX = is_to_box(Path(TOP_SRC), STOP_STEP)

    gus.set_free_parameters(user_input.ask_parameters(TOP_SRC, TOP_MODULE))

    # A local copy of the top module source file is needed
    # in order to update the parameters in place
    if not TO_BOX:
        with Path(
            SRC_FOLDER + get_config("VHDL_LOCAL_SRC")
            if TOP_SUFFIX == ".vhd"
            else SRC_FOLDER + get_config("VERILOG_LOCAL_SRC")
        ).open("w+") as f:
            f.write(Path(TOP_SRC).read_text())

    # Clock and out are needed for boxing the component
    # to avoid overflowing pins
    # And to deal with incorrect reverse parsing from the HDLConv library
    # when dealing with (System)Verilog source code or some VHDL source code
    else:
        CLOCK_PORT = user_input.ask_identifiers(TOP_SRC, TOP_MODULE)
        frame.fill_box(
            get_config("VHDL_DIR") + get_config("VHDL_BOX_FRAME")
            if TOP_SUFFIX == ".vhd"
            else get_config("VERILOG_DIR") + get_config("VERILOG_BOX_FRAME"),
            TOP_SRC,
            TOP_MODULE,
            gus.FREE_PARAMETERS,
            CLOCK_PORT,
            get_config("PLACEHOLDER"),
            get_config("VHDL_DIR") + get_config("VHDL_BOX")
            if TOP_SUFFIX == ".vhd"
            else get_config("VERILOG_DIR") + get_config("VERILOG_BOX"),
        )

    INCREMENTAL_MODE = user_input.ask_incremental_mode(STOP_STEP)

    # Directives to give to the -directive switch in
    # synth_design and eventually in place_design and route_design
    # some directives are forbidden in incremental mode
    # for place_design and route_design
    (
        SYNTHESIS_DIRECTIVE,
        PLACE_DIRECTIVE,
        ROUTE_DIRECTIVE,
    ) = user_input.ask_directives(STOP_STEP, INCREMENTAL_MODE)

    # Clock given in Mhz
    TARGET_CLOCK = user_input.ask_clock()

    # Swapping out placeholder in constraint file
    frame.fill(
        get_config("XDC_DIR") + get_config("CONSTRAINT_FRAME"),
        [str(1000 / TARGET_CLOCK)],
        get_config("PLACEHOLDER"),
        get_config("XDC_DIR") + get_config("CONSTRAINT"),
    )
    frame.fill_tcl(
        SRC_FOLDER,
        TOP_SRC,
        TOP_MODULE,
        TO_BOX,
        SYNTHESIS_PART,
        SYNTHESIS_DIRECTIVE,
        INCREMENTAL_MODE,
        STOP_STEP,
        TOP_LANG,
        TARGET_CLOCK,
        PLACE_DIRECTIVE,
        ROUTE_DIRECTIVE,
    )

    pe.setup_evaluation(
        STOP_STEP,
        TOP_LANG,
        TO_BOX,
        TOP_MODULE,
        SRC_FOLDER,
        TOP_SRC,
        TARGET_CLOCK,
        INCREMENTAL_MODE,
    )
    if user_input.ask_is_point_evaluation():
        print(
            pe.evaluate(
                tuple(user_input.ask_parameters_value(gus.FREE_PARAMETERS))
            )
        )

    else:
        gus.set_free_parameters_range(
            user_input.ask_parameters_range(gus.FREE_PARAMETERS)
        )
        es.generate_dataset(
            get_config("INITIAL_SAMPLES"),
            gus.FREE_PARAMETERS_RANGE,
            gus.FREE_PARAMETERS,
        )
        fit.set_threshold(es.examples)

        result = optimize(get_config("GENETIC_RUN_TIME"))
        print("Optimization Result: " + str(result))



from functools import lru_cache
from pathlib import Path
from dataclasses import dataclass
from typing import Tuple, Dict
from dovado_rtl.config import get_config
import dovado_rtl.vivado_interaction as vivado
import dovado_rtl.report_parsing as report
import dovado_rtl.src_parsing as src
import dovado_rtl.global_user_selections as gus
import dovado_rtl.user_input as user_input
import numpy as np
from dovado_rtl.frame_handling import setup_incremental
from dovado_rtl.src_parsing import IsIncremental


@dataclass
class EvaluationSetup:
    stop_step: src.StopStep
    top_lang: src.RTL
    is_boxed: bool
    top_module: str
    src_folder: str
    top_src: str
    target_clock: float
    is_incremental: IsIncremental


@dataclass
class DesignValue:
    utilisation: Dict[Tuple[str, str], float]
    negative_max_frequency: float


evaluation_setup = None
is_evaluation_setup = False


def setup_evaluation(
    stop_step,
    top_lang,
    to_box,
    top_module,
    src_folder,
    top_src,
    target_clock,
    is_incremental,
):
    global is_evaluation_setup
    if is_evaluation_setup:
        return
    global evaluation_setup
    evaluation_setup = EvaluationSetup(
        stop_step,
        top_lang,
        to_box,
        top_module,
        src_folder,
        top_src,
        target_clock,
        is_incremental,
    )
    is_evaluation_setup = True
    print("Evaluation Setup: " + str(evaluation_setup))


is_first_evaluation = True


@lru_cache
def evaluate(design_point: Tuple[int]) -> DesignValue:
    global is_first_evaluation
    if not evaluation_setup.is_boxed:
        for parameter, value in zip(gus.FREE_PARAMETERS, design_point):
            src.set_parameter(
                Path.joinpath(
                    Path(evaluation_setup.src_folder),
                    get_config("VHDL_LOCAL_SRC"),
                ),
                evaluation_setup.top_module,
                parameter,
                value,
            )
    else:
        for parameter, value in zip(gus.FREE_PARAMETERS, design_point):
            src.map_parameter(
                Path(get_config("VHDL_DIR") + get_config("VHDL_BOX"))
                if evaluation_setup.top_lang is src.RTL.VHDL
                else Path(
                    get_config("VERILOG_DIR") + get_config("VERILOG_BOX")
                ),
                "box",
                parameter,
                value,
            )
    vivado_out, success = vivado.source(
        get_config("TCL_DIR") + get_config(evaluation_setup.stop_step.name)
    )
    print(vivado_out)
    if is_first_evaluation:
        is_first_evaluation = False
        setup_incremental(evaluation_setup.is_incremental)
    if not gus.METRICS:
        gus.set_metrics(
            user_input.ask_utilization_metrics(
                report.get_available_indices(
                    get_config("VIVADO_OUTPUT_DIR")
                    + get_config(
                        evaluation_setup.stop_step.name + "_UTILISATION"
                    )
                )
            )
        )

    return (
        DesignValue(
            utilisation={i: np.inf for i in gus.METRICS},
            negative_max_frequency=np.inf,
        )
        if not success
        else DesignValue(
            utilisation={
                i: report.get_utilisation(
                    get_config("VIVADO_OUTPUT_DIR")
                    + get_config(
                        evaluation_setup.stop_step.name + "_UTILISATION"
                    ),
                    i[0],
                    i[1],
                )
                for i in gus.METRICS
            },
            negative_max_frequency=-float(
                get_max_frequency(
                    report.get_wns(
                        get_config("VIVADO_OUTPUT_DIR")
                        + get_config(
                            evaluation_setup.stop_step.name + "_TIMING"
                        )
                    )
                )
            ),
        )
    )


def get_max_frequency(wns):

    return 1000 / ((1 / 1000 * evaluation_setup.target_clock) - wns)


def get_metric(design_value, metric):
    if metric == "max_frequency":
        return design_value.negative_max_frequency
    else:
        try:
            return design_value.utilisation[metric]
        except Exception as e:
            raise Exception(
                "Invalid metric passed to the fitness function: "
                + str(metric)
                + "\n"
                + str(e)
            )



import re
from pathlib import Path
import dovado_rtl.src_parsing as parsing
from dovado_rtl.config import get_config


def fill(frame_path, replacements, placeholder, out_path):
    out = re.sub(
        placeholder,
        lambda m, r=iter(replacements): next(r),
        Path(frame_path).open("r").read(),
    )
    Path(out_path).write_text(out)


def fill_tcl(
    src_folder,
    top_src,
    top_module,
    is_boxed,
    synthesis_part,
    synthesis_directive,
    incremental_mode,
    stop_step,
    top_lang,
    target_clock,
    place_directive="",
    route_directive="",
):
    print("TOP LANG: " + str(top_lang))
    SYNTHESIS_REPLACEMENTS = [
        get_config("VIVADO_OUTPUT_DIR"),
        src_folder,
        get_config("XDC_DIR") + get_config("CONSTRAINT"),
        (
            "read_vhdl -library bftLib "
            + get_config("VHDL_DIR")
            + get_config("VHDL_BOX")
        )
        if top_lang is parsing.RTL.VHDL
        else (
            "read_verilog "
            + get_config("VERILOG_DIR")
            + get_config("VERILOG_BOX")
        ),
        ("set_property IS_ENABLED 0 [get_files " + top_src + "]")
        if top_lang is parsing.RTL.VHDL
        else "# no disabling needed",
        top_module if not is_boxed else "box",
        synthesis_part,
        synthesis_directive,
        "-incremental_synth" if incremental_mode.synthesis else "",
    ]
    print("SYNTHESIS REPLACEMENTS: " + str(SYNTHESIS_REPLACEMENTS))

    IMPLEMENTATION_REPLACEMENTS = (
        []
        if stop_step is parsing.StopStep.SYNTHESIS
        else (
            [
                "",
                "-directive " + place_directive,
                "",
                "-directive " + route_directive,
            ]
            if not incremental_mode.implementation
            else [
                "-directive " + place_directive,
                " ",
                "-directive " + route_directive,
                " ",
            ]
        )
    )

    REPLACEMENT = (
        SYNTHESIS_REPLACEMENTS
        if stop_step is parsing.StopStep.SYNTHESIS
        else SYNTHESIS_REPLACEMENTS[:3]
        + [
            (
                "read_vhdl -library bftLib "
                + get_config("VHDL_DIR")
                + get_config("VHDL_BOX")
            )
            if top_lang is parsing.RTL.VHDL
            else (
                "read_verilog "
                + get_config("VERILOG_DIR")
                + get_config("VERILOG_BOX")
            )
        ]
        + SYNTHESIS_REPLACEMENTS[5:]
        + IMPLEMENTATION_REPLACEMENTS
    )

    print("REPLACEMENTS: " + str(REPLACEMENT))
    fill(
        get_config("TCL_DIR") + get_config(stop_step.name + "_FRAME"),
        REPLACEMENT
        + [
            get_config(stop_step.name + "_TIMING"),
            get_config(stop_step.name + "_UTILISATION"),
        ],
        get_config("PLACEHOLDER"),
        get_config("TCL_DIR") + get_config(stop_step.name),
    )


def fill_box(
    frame_path,
    top_src,
    top_module,
    parameters,
    clock_port,
    placeholder,
    out_path,
):
    full_parameters = []
    for param in parameters:
        full_parameters.append(
            parsing.get_parameter_on_name(Path(top_src), top_module, param)
        )

    top_suffix = Path(top_src).suffix
    if top_suffix == ".vhd":
        _vhdl_fill_box(
            frame_path,
            top_src,
            top_module,
            full_parameters,
            clock_port,
            placeholder,
            out_path,
        )
    elif top_suffix == ".sv" or top_suffix == ".v":
        _verilog_fill_box(
            frame_path,
            top_src,
            top_module,
            full_parameters,
            clock_port,
            placeholder,
            out_path,
        )
    else:
        raise ValueError(
            "Suffix of both Source and Frame is invalid: " + top_suffix
        )


def _vhdl_parameter_map(parameters):
    parameter_section = "generic map(\n"
    for parameter in parameters[:-1]:
        if not parameter.value:
            raise Exception(
                "Please provide default values for all parameters, "
                + str(parameter.name)
                + " does not have one."
            )
        try:
            parameter_section += (
                parameter.name
                + " => "
                + (
                    str(parameter.value)
                    if not parameter.value.base
                    else str(
                        int(parameter.value.val[1:], parameter.value.base,)
                    )
                ).strip()
                + ",\n"
            )
        except Exception:
            print("Skipping unsupported parameter: " + str(parameter))
            continue
    try:
        parameter_section += (
            parameters[-1].name
            + " => "
            + (
                str(parameters[-1].value)
                if not parameters[-1].value.base
                else str(
                    int(
                        parameters[-1].value.val[1:],
                        parameters[-1].value.base,
                    )
                )
            )
            + ")"
        )
    except Exception:
        print("Skipping unsupported parameter: " + str(parameter))
        parameter_section += ")"
    return parameter_section


def _vhdl_fill_box(
    frame_path,
    top_src,
    top_module,
    parameters,
    clock_port,
    placeholder,
    out_path,
):
    libraries, imports = parsing.get_imports(Path(top_src))
    input_mapping = [
        parsing.get_port_id(port)
        + (
            " => '1',\n"
            if parsing.get_port_type(port) == "std_logic"
            else " => " + parsing.get_port_type(port) + "'((others => '1')),\n"
        )
        for port in parsing.get_ports(Path(top_src), top_module)
        if (
            parsing.get_port_direction(port) == "IN"
            and parsing.get_port_id(port) != parsing.get_port_id(clock_port)
        )
    ]
    input_mapping[len(input_mapping) - 1] = input_mapping[
        len(input_mapping) - 1
    ].replace(",", "")
    replacements = [
        "".join(
            ["library " + lib + ";\n" for lib in libraries]
            + ["use " + imp + ";\n" for imp in imports]
        ),
        "Work." + top_module,
        _vhdl_parameter_map(parameters),
        parsing.get_port_id(clock_port),
        "".join(input_mapping),
    ]
    fill(frame_path, replacements, placeholder, out_path)


def _verilog_parameter_map(parameters):
    parameter_section = "#(\n"
    for parameter in parameters[:-1]:
        parameter_section += (
            "."
            + parameter.name
            + "("
            + str(int(str(parameter.value.val), parameter.value.base))
            + "),\n"
        )
    parameter_section += (
        "."
        + parameters[-1].name
        + "("
        + str(int(str(parameters[-1].value.val), parameters[-1].value.base,))
        + ")\n)\n"
    )
    return parameter_section


def _verilog_fill_box(
    frame_path,
    top_src,
    top_module,
    parameters,
    clock_port,
    placeholder,
    out_path,
):
    input_mapping = [
        "." + parsing.get_port_id(port) + " ('1),\n"
        for port in parsing.get_ports(Path(top_src), top_module)
        if (
            parsing.get_port_direction(port) == "IN"
            and parsing.get_port_id(port) != parsing.get_port_id(clock_port)
        )
    ]

    input_mapping[len(input_mapping) - 1] = input_mapping[
        len(input_mapping) - 1
    ].replace(",", "")

    replacements = [
        top_module,
        _verilog_parameter_map(parameters),
        parsing.get_port_id(clock_port),
        "".join(input_mapping),
    ]

    fill(frame_path, replacements, placeholder, out_path)


def setup_incremental(is_incremental: parsing.IsIncremental):
    if not is_incremental.synthesis and not is_incremental.implementation:
        return
    if is_incremental.synthesis:
        path = get_config("TCL_DIR") + get_config("SYNTHESIS")
        synthesis = Path(path).read_text()
        synthesis = synthesis.replace("#! ", "")
        Path(path).write_text(synthesis)
    if is_incremental.implementation:
        path = get_config("TCL_DIR") + get_config("IMPLEMENTATION")
        implementation = Path(path).read_text()
        implementation = implementation.replace("#! ", "")
        Path(path).write_text(implementation)



from typing import List, Dict, Tuple

FREE_PARAMETERS_RANGE: Dict[str, Tuple[int, int]] = None
METRICS: List[Tuple[str, str]] = None
FREE_PARAMETERS: List[str] = None


def set_free_parameters_range(free_parameters_range):
    global FREE_PARAMETERS_RANGE
    FREE_PARAMETERS_RANGE = free_parameters_range


def set_metrics(metrics):
    global METRICS
    METRICS = metrics


def set_free_parameters(free_parameters):
    global FREE_PARAMETERS
    FREE_PARAMETERS = free_parameters
    print("FREE PARAM: " + str(FREE_PARAMETERS))



from pymoo.model.problem import Problem
from pymoo.algorithms.nsga2 import NSGA2
from pymoo.factory import (
    get_sampling,
    get_crossover,
    get_mutation,
    get_termination,
)
from pymoo.optimize import minimize

import numpy as np
import dovado_rtl.global_user_selections as gus
from dovado_rtl.fitness import fitness
import pickle


class MyProblem(Problem):
    def __init__(self):
        super().__init__(
            n_var=len(gus.FREE_PARAMETERS_RANGE),
            n_obj=len(gus.METRICS),
            n_constr=0,
            xl=[
                gus.FREE_PARAMETERS_RANGE[parameter][0]
                for parameter in gus.FREE_PARAMETERS
            ],
            xu=[
                gus.FREE_PARAMETERS_RANGE[parameter][1]
                for parameter in gus.FREE_PARAMETERS
            ],
            type_var=np.int,
            elementwise_evaluation=True,
        )

    def _evaluate(self, x, out, *args, **kwargs):
        out["F"] = np.column_stack(
            [
                fitness(tuple(x), metric)
                for metric in (gus.METRICS + ["max_frequency"])
            ]
        )


def optimize(execution_time: str) -> np.ndarray:
    problem = MyProblem()

    algorithm = NSGA2(
        pop_size=10 * len(gus.FREE_PARAMETERS),
        n_offsprings=10 * len(gus.FREE_PARAMETERS),
        sampling=get_sampling("int_random"),
        crossover=get_crossover("int_sbx", prob=0.9, eta=15),
        mutation=get_mutation("int_pm", eta=20),
        eliminate_duplicates=True,
    )

    termination = get_termination("time", execution_time)
    res = minimize(
        problem,
        algorithm,
        termination,
        seed=1,
        save_history=True,
        verbose=True,
    )
    with open("objs.pkl", "wb") as f:
        pickle.dump([res, algorithm], f)
    # with open('objs.pkl', 'rb') as f:
    #     obj0, obj1, obj2 = pickle.load(f)
    return res.F[-1]



from pathlib import Path
import yaml

_config_dict = {
    # Directory holding tcl scripts
    "TCL_DIR": "tcl/",
    # Directory holding xdc files
    "XDC_DIR": "xdc/",
    # Directory holding vhdl files
    "VHDL_DIR": "vhdl/",
    # Directory holding verilog files
    "VERILOG_DIR": "verilog/",
    # Constraint file for setting master clock frequency
    # with blank clock to be filled
    "CONSTRAINT_FRAME": "constraint_frame.xdc",
    # Constraint file with clock filled
    "CONSTRAINT": "constraint.xdc",
    # TCL script with blank fields to be filled
    # with user input in order to evaluate vhdl source until synthesis
    "SYNTHESIS_FRAME": "synthesis_frame.tcl",
    # TCL script with all fields filled with user input until synthesis
    "SYNTHESIS": "synthesis.tcl",
    # TCL script with blank fields to be filled
    # with user input in order to evaluate vhdl source
    "IMPLEMENTATION_FRAME": "implementation_frame.tcl",
    # TCL script with all fields filled with user input
    "IMPLEMENTATION": "implementation.tcl",
    # Local copy of VHDL top module source where parameters are changed in place.
    # This is used if stop step is synthesis
    # Reverse parsing does not work for (System)Verilog so no such file
    # exists for the (System)Verilog case which is instead always boxed
    "VHDL_LOCAL_SRC": "dovado_work_file.vhd",
    # Vhdl file with fields to be filled for boxing
    "VHDL_BOX_FRAME": "box_frame.vhd",
    # Vhdl file with all fields filled
    "VHDL_BOX": "box.vhd",
    # TODO remove this or understand its possible role
    # Verilog local source
    "VERILOG_LOCAL_SRC": "dovado_work_file.sv",
    # Verilog file with fields to be filled for boxing
    "VERILOG_BOX_FRAME": "box_frame.sv",
    # Verilog file with all fields filled
    "VERILOG_BOX": "box.sv",
    # Placeholder inside the two above files
    # to be swapped with user input
    "PLACEHOLDER": "____",
    # Output of vivado call (reports and checkpoints)
    "VIVADO_OUTPUT_DIR": "vivado_out/",
    # After Synthesis Utilisation Report Name
    "SYNTHESIS_UTILISATION": "post_synth_util.xml",
    # After Implementation Utilisation Report Name
    "IMPLEMENTATION_UTILISATION": "post_impl_util.xml",
    # After Synthesis Timing (setup) Report Name
    "SYNTHESIS_TIMING": "post_synth_setup_timing.rpt",
    # After Implementation Timing (setup) Report Name
    "IMPLEMENTATION_TIMING": "post_impl_setup_timing.rpt",
    # Maximum integer value considered for RTL parameters
    "INTEGER_HIGH": 2147483647,
    # Initial sample size for the estimator
    "INITIAL_SAMPLES": 100,
    # N-parameter to choose which example to calculate the distance from
    "N": 3,
    # Run time set as a constraint for the genetic algorithm, format "hh:mm:ss"
    "GENETIC_RUN_TIME": "00:30:00",
}

custom_configuration_checked = False
configuration_exists = False


def get_config(config_name: str):
    global custom_configuration_checked
    global configuration_exists
    if not custom_configuration_checked:
        custom_configuration_checked = True
        try:
            CONFIG = yaml.safe_load(Path("config.yaml").open())
            configuration_exists = True
        except:
            configuration_exists = False
    if configuration_exists:
        try:
            return CONFIG[config_name]
        except:
            return _config_dict[config_name]
    else:
        return _config_dict[config_name]



import math
from pathlib import Path
from functools import lru_cache
from heapq import nlargest
import dovado_rtl.point_evaluation as pe
import dovado_rtl.estimation as es
from dovado_rtl.config import get_config
import numpy as np
from typing import Tuple, List


threshold: int = 0

estimate_working = True


@lru_cache
def fitness(design_point: Tuple[int], metric: Tuple[str, str]):
    # design_point is a Tuple because lists are unhashable
    # while caching is allowed only with hashable parameters
    global estimate_working
    print(
        "Nth nearest distance: "
        + str(
            _nth_nearest_distance(design_point, es.examples, get_config("N"))
        )
    )
    if (
        (
            _nth_nearest_distance(design_point, es.examples, get_config("N"))
            == 0
        )
        or (
            _nth_nearest_distance(design_point, es.examples, get_config("N"))
            > threshold
        )
        or (not estimate_working)
    ):
        print("Fitness calling Vivado directly")
        full_design_value = pe.evaluate(design_point)
        design_value = pe.get_metric(full_design_value, metric)
        if (
            _nth_nearest_distance(design_point, es.examples, get_config("N"))
            > threshold
        ) and estimate_working:

            if any(
                np.isinf(v) for v in full_design_value.utilisation.values()
            ) or np.isinf(full_design_value.negative_max_frequency):
                es.add_example(
                    es.Example(design_point, full_design_value, True)
                )
            else:
                es.add_example(
                    es.Example(design_point, full_design_value, False)
                )
            set_threshold(es.examples)
    else:
        print("Fitness estimating")
        design_value = es.estimate(design_point, metric)
        if not design_value:
            print("Disabling estimator")
            estimate_working = False
            full_design_value = pe.evaluate(design_point)
            design_value = pe.get_metric(full_design_value, metric)

    print("design_point: " + str(design_point))
    print("metric: " + str(metric))
    print("Design Value: " + str(design_value))
    return design_value


def _distance(design_point_1: List[int], design_point_2: List[int]) -> float:
    if len(design_point_1) != len(design_point_2):
        raise ValueError("Mismatching design points: different dimensionality")
    return math.sqrt(
        np.sum(
            np.power(
                np.subtract(
                    np.array(design_point_1), np.array(design_point_2)
                ),
                np.array([2] * len(design_point_1)),
            )
        )
        / len(design_point_1)
    )


def _nth_nearest_distance(
    design_point: List[int], examples: List[es.Example], n: int,
) -> float:
    distances = []
    for example in examples:
        distances.append(_distance(design_point, example.design_point))

    try:
        return nlargest(n, distances)[n - 1]
    except Exception:
        return nlargest(n, distances)[-2]


def _mean(numbers):
    return float(sum(numbers)) / max(len(numbers), 1)


def set_threshold(examples):
    global threshold
    distances = []
    for example in examples:
        distances.append(
            _nth_nearest_distance(
                example.design_point, examples, get_config("N")
            )
        )
    print("Distances for threshold: " + str(distances))
    threshold = int(round(_mean(distances)))
    print("Set Threshold: " + str(threshold))



from hdlConvertorAst.language import Language
from hdlConvertor import HdlConvertor
from hdlConvertorAst.hdlAst import (
    HdlModuleDef,
    HdlModuleDec,
    HdlImport,
    HdlAll,
    HdlValueInt,
    HdlCompInst,
    HdlOp,
    HdlValueId,
)
from hdlConvertor._hdlConvertor import ParseException
from hdlConvertorAst.to.vhdl.vhdl2008 import ToVhdl2008
from hdlConvertorAst.to.verilog.verilog2005 import ToVerilog2005
import io
import yaml
from pathlib import Path
from enum import Enum, auto
from dataclasses import dataclass


class RTL(Enum):
    VHDL = auto()
    VERILOG = auto()
    SYSTEM_VERILOG = auto()


class StopStep(Enum):
    SYNTHESIS = auto()
    IMPLEMENTATION = auto()


class ImplementationStep(Enum):
    PLACE = auto()
    ROUTE = auto()


@dataclass
class IsIncremental:
    synthesis: bool = True
    implementation: bool = True


CONFIG = yaml.safe_load(Path("config.yaml").open())

parsed_src_path = None
parsed = None


def parse(src_path):
    global parsed_src_path
    global parsed
    parsed_src_path = src_path
    src_path_string = str(src_path)
    if src_path.suffix == ".vhd":
        try:
            parsed = HdlConvertor().parse(
                src_path_string,
                Language.VHDL,
                [],
                hierarchyOnly=False,
                debug=True,
            )
        except ParseException as e:
            raise ValueError(
                "Could not parse " + src_path_string + " as VHDL\n" + str(e)
            )
    elif src_path.suffix == ".v":
        try:
            parsed = HdlConvertor().parse(
                src_path_string,
                Language.VERILOG,
                [],
                hierarchyOnly=False,
                debug=True,
            )
        except ParseException as e:
            raise ValueError(
                "Could not parse "
                + str(src_path)
                + " as Verilog, .v files must be Verilog,"
                + " SystemVerilog files must be .sv\n"
                + str(e)
            )
    elif src_path.suffix == ".sv":
        try:
            parsed = HdlConvertor().parse(
                src_path_string,
                Language.SYSTEM_VERILOG,
                [],
                hierarchyOnly=False,
                debug=True,
            )
        except ParseException:
            raise ValueError(
                "Could not parse "
                + str(src_path)
                + " as SystemVerilog, .sv files must be SystemVerilog,"
                + " Verilog files must be .v"
            )


def _is_parsed(src_path):
    if not parsed or parsed_src_path != src_path:
        return False
    return True


def is_to_box(src_path, stop_step):
    if stop_step is not StopStep.SYNTHESIS:
        return True
    if src_path.suffix == ".sv" or ".v":
        return True
    if not _is_parsed(src_path):
        parse(src_path)
    out_path = "vhdl/local_test_copy.vhd"
    write_to_file(out_path)
    try:
        parse(out_path)
        return False
    except Exception:
        return False


def get_modules(src_path):
    if not _is_parsed(src_path):
        parse(src_path)
    modules = []
    for o in parsed.objs:
        if isinstance(o, HdlModuleDec):
            modules.append(o.name)
        if isinstance(o, HdlModuleDef):
            if o.dec:
                modules.append(o.dec.name)
    return modules


def get_parameters(src_path, module):
    if not _is_parsed(src_path):
        parse(src_path)
    for o in parsed.objs:
        if isinstance(o, HdlModuleDec) and o.name == module:
            return o.params
        if isinstance(o, HdlModuleDef) and o.dec and o.dec.name == module:
            return o.dec.params


def get_parameter_range(parameter):
    # this method only works with VHDL parameters
    if isinstance(parameter.type, HdlOp):
        # case in which a range is specified
        if parameter.type.ops[1].fn.name != "TO":
            raise Exception(
                "Only 'to' is supported as an operation"
                + " for parameter range specification, input: "
                + str(parameter.type.ops[1].fn)
            )
        return tuple(parameter.type.ops[1].ops)
    try:
        return {
            "integer": (-CONFIG["INTEGER_HIGH"], CONFIG["INTEGER_HIGH"]),
            "natural": (0, CONFIG["INTEGER_HIGH"]),
            "positive": (1, CONFIG["INTEGER_HIGH"]),
            "signed": (-CONFIG["INTEGER_HIGH"], CONFIG["INTEGER_HIGH"]),
            "unsigned": (0, CONFIG["INTEGER_HIGH"]),
        }[str(parameter.type).lower()]
    except Exception as e:
        raise Exception(
            "Unsupported type: " + str(parameter.type).lower() + "\n" + str(e)
        )


added = False


def _add_dont_touch():
    global added
    if added:
        return
    for o in parsed.objs:
        if isinstance(o, HdlModuleDef):
            for def_o in o.objs:
                if isinstance(def_o, HdlCompInst):
                    def_o.module_name = HdlValueId(
                        '(* dont_touch  = "true" *) ' + str(def_o.module_name)
                    )
                    added = True
                    return


def write_to_file(out_path, no_touch=False):
    with open(out_path, "w") as f:
        if out_path.suffix == ".vhd":
            # explicit conversion of all instantiations to direct instantiations
            # mandatory for Vivado compatibility
            for o in parsed.objs:
                if isinstance(o, HdlModuleDef):
                    for def_obj in o.objs:
                        if isinstance(def_obj, HdlCompInst) and isinstance(
                            def_obj.module_name, HdlOp
                        ):
                            def_obj.module_name.ops[0] = HdlValueId(
                                "entity work"
                            )

            if not no_touch:
                ToVhdl2008(f).visit_HdlContext(parsed)
            else:
                stream = io.StringIO()
                ToVhdl2008(stream).visit_HdlContext(parsed)
                lines = [line for line in stream.getvalue().split("\n")]
                index = None
                for line in lines:
                    if line.lower() == "architecture box_arch of box is":
                        index = lines.index(line)
                if index:
                    lines.insert(index + 1, "attribute DONT_TOUCH : string;")
                    lines.insert(
                        index + 2,
                        'attribute DONT_TOUCH of BOXED : label is "TRUE";',
                    )
                f.write("\n".join(lines))
        else:
            if no_touch:
                _add_dont_touch()
            ToVerilog2005(f).visit_HdlContext(parsed)
    global parsed_src_path
    parsed_src_path = out_path


def get_parameter_on_name(src_path, module, parameter):
    if not _is_parsed(src_path):
        parse(src_path)
    for o in parsed.objs:
        if isinstance(o, HdlModuleDec) and o.name == module:
            for param in o.params:
                if param.name == parameter:
                    return param

        if isinstance(o, HdlModuleDef) and o.dec and o.dec.name == module:
            for param in o.dec.params:
                if param.name == parameter:
                    return param


def set_parameter(src_path, module, parameter, value):
    # value must be base 10
    if not _is_parsed(src_path):
        parse(src_path)
    for o in parsed.objs:
        if isinstance(o, HdlModuleDec) and o.name == module:
            for param in o.params:
                if param.name == parameter.name:
                    param.value = HdlValueInt(value, None, None)

        if isinstance(o, HdlModuleDef) and o.dec and o.dec.name == module:
            for param in o.dec.params:
                if param.name == parameter.name:
                    param.value = HdlValueInt(value, None, None)

    write_to_file(src_path)


def map_parameter(src_path, module, parameter, value):
    print(
        "MAPPING PARAMETER: "
        + str(src_path)
        + "  "
        + str(module)
        + "  "
        + str(parameter)
        + "  "
        + str(value)
    )
    if not _is_parsed(src_path):
        parse(src_path)
    for o in parsed.objs:
        if isinstance(o, HdlModuleDef):
            for module_definition in o.objs:
                if (
                    isinstance(module_definition, HdlCompInst)
                    and module_definition.param_map
                ):
                    for param in module_definition.param_map:
                        if str(param.ops[0]) == parameter:
                            param.ops[1] = HdlValueInt(value, None, None)
                            write_to_file(src_path, no_touch=True)
                            return
        if isinstance(o, HdlModuleDec):
            for module_definition in o.objs:
                if (
                    isinstance(module_definition, HdlCompInst)
                    and module_definition.param_map
                ):
                    for param in module_definition.param_map:
                        if str(param.ops[0]) == parameter:
                            param.ops[1] = HdlValueInt(value, None, None)
                            write_to_file(src_path, no_touch=True)
                            return


def get_ports(src_path, module):
    if not _is_parsed(src_path):
        parse(src_path)
    for o in parsed.objs:
        if isinstance(o, HdlModuleDec) and o.name == module:
            return o.ports
        if isinstance(o, HdlModuleDef) and o.dec and o.dec.name == module:
            return o.dec.ports


def get_port_id(port):
    return str(port.name)


def get_port_from_id(port, ports):
    for i in ports:
        if i.name == port:
            return i
    return None


def get_port_direction(port):
    direction = str(port.direction)
    return direction[direction.find(".") + 1 :]


def get_port_type(port):
    try:
        return str(port.type.ops[0])
    except AttributeError:
        return str(port.type)


def get_imports(src_path):
    if not _is_parsed(src_path):
        parse(src_path)
    libraries = []
    use_clauses = []
    for o in parsed.objs:
        if isinstance(o, HdlImport):
            if o.path[0].val not in libraries and o.path[0].val != "work":
                libraries.append(o.path[0].val)
            use_clause = o.path[0].val + "." + o.path[1].val
            if HdlAll in o.path:
                use_clause += ".all"
            if use_clause not in use_clauses:
                use_clauses.append(use_clause)
    return libraries, use_clauses
